'''Main module for the FASTA file processing tool'''
import argparse
import os

from packages.utils.seq_file_manager import SeqFileManager
from packages.plots.plot_generator import PlotGenerator
from packages.sequences.sequences import Sequence
from packages.transformers.duplicated_transformer import DuplicatedTransformer
from packages.transformers.reverse_complement import ReverseComplement
from packages.stats.sequence_stats import SequenceStats

def main():
    '''Main function'''
    parser = argparse.ArgumentParser(description='FASTA file processing tool')
    parser.add_argument('--dremove', action='store_true', help='Remove duplicates')
    parser.add_argument('--drename', action='store_true', help='Rename duplicates')
    parser.add_argument('--reverse', action='store_true', help='Reverse sequences')
    parser.add_argument('--complement', action='store_true', help='Complement sequences')
    parser.add_argument('--rc', action='store_true', help='Reverse complement sequences')
    parser.add_argument('--stats', help='Compute stats', action='store_true')
    parser.add_argument('--casefile', help='Case transformation (original, upper, lower)')
    parser.add_argument('--plots', help='Generate plots', action='store_true')
    args = parser.parse_args()


    # Get the FASTA files in the current directory
    current_directory = os.getcwd()
    fasta_files = [file for file in os.listdir(current_directory) if file.endswith('.fasta')]

    # Verify that there are FASTA files in the current directory
    if not any([args.casefile, args.dremove, args.drename, args.reverse, args.complement, args.rc, args.stats, args.plots]):
        print('No transformation specified')
        return

    if not fasta_files:
        print('No FASTA files found in the current directory')
        return
    
    stats_directory = os.path.join(current_directory, 'result_stats')

    for filename in fasta_files:
        seq_file_manager = SeqFileManager(case=args.casefile, input_filename=filename)
        sequences = seq_file_manager.load_fasta(filename)

        # Apply case transformation if specified
        if args.casefile == 'upper':
            sequences = [Sequence(seq.id, seq.seq.upper(), seq.file_name) for seq in sequences]
        elif args.casefile == 'lower':
            sequences = [Sequence(seq.id, seq.seq.lower(), seq.file_name) for seq in sequences]

        if args.dremove:
            # Remove duplicates
            sequences = DuplicatedTransformer(style='remove').transform(sequences)
        elif args.drename:
            # Rename duplicates
            sequences = DuplicatedTransformer(style='rename').transform(sequences)

        if args.reverse:
            # Reverse sequences
            sequences = ReverseComplement(style='reverse').transform(sequences)

        elif args.complement:
            # Complement sequences
            sequences = ReverseComplement(style='complement').transform(sequences)

        elif args.rc:
            # Reverse complement sequences
            sequences = ReverseComplement(style='both').transform(sequences)

        if args.plots:
            plot_generator = PlotGenerator(sequences, filename, 'result_plots')
            plot_generator.generate_plots()

        if args.stats:
            sequence_stats = SequenceStats()
            sequence_stats.generate_stats(sequences, stats_directory, filename)

        # Process the sequences if any transformation was specified
        if any([args.casefile, args.dremove, args.drename, args.reverse, args.complement, args.rc]):
            output_directory = os.path.join(current_directory, 'result_formatted')
            seq_file_manager.write_fasta(sequences, output_directory)

if __name__ == '__main__':
    main()



'''Module to compute statistics for a list of sequences'''
import csv
import os

class SequenceStats:
    '''Class to compute statistics for a list of sequences'''
    def get_seq_stats(self, sequence):
        '''Computes statistics for a sequence'''
        # Calculates statistics for an individual sequence
        length = len(sequence.seq)
        counts = {'A': 0, 'C': 0, 'T': 0, 'G': 0, '-': 0}

        for base in sequence.seq:
            if base in counts:
                counts[base] += 1

        # Return statistics in a list
        stats = [sequence.id, length] + [counts[base] for base in 'ACTG-']
        return stats

    def get_seq_stats_matrix(self, seq):
        '''Computes statistics for a list of sequences'''
        header = ['id', 'len', 'A', 'C', 'T', 'G', '-']
        matrix = [header]

        for sequence in seq:
            stats = SequenceStats().get_seq_stats(sequence)
            matrix.append(stats)

        # Returns the statistics array
        return matrix

    def stats_to_csv(self, matrix, filepath):
        '''Writes the statistics array to a CSV file'''
        with open(filepath, 'w', newline='', encoding='utf-8') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerows(matrix)

    def generate_stats(self, sequences, stats_directory, filename):
        '''Generates statistics for a list of sequences'''
        os.makedirs(stats_directory, exist_ok=True)
        stats_matrix = self.get_seq_stats_matrix(sequences)
        stats_filename = os.path.splitext(filename)[0] + '_stats.csv'
        stats_filepath = os.path.join(stats_directory, stats_filename)
        self.stats_to_csv(stats_matrix, stats_filepath)



'''Returns a list of Sequence objects with the specified transformation.'''
# pylint: disable=too-few-public-methods
from .abstract_transformer import AbstractTransformer

class DuplicatedTransformer(AbstractTransformer):
    '''Returns a list of Sequence objects with the specified transformation.'''
    def __init__(self, style):
        self.style = style

    def transform(self, sequence):
        new_seq_list = []

        if self.style == 'rename':
            id_dict = {}
            duplicate_count = {}
            for seq in sequence:
                seq_id = seq.id
                if seq_id in id_dict:
                    if seq_id not in duplicate_count:
                        duplicate_count[seq_id] = 1
                        id_dict[seq_id].id = f'{seq_id}.{duplicate_count[seq_id]}'
                        seq.id = f'{seq_id}.{duplicate_count[seq_id] + 1}'
                        id_dict[seq.id] = seq
                        duplicate_count[seq_id] += 1
                    else:
                        duplicate_count[seq_id] += 1
                        seq.id = f'{seq_id}.{duplicate_count[seq_id]}'
                        id_dict[seq.id] = seq
                else:
                    id_dict[seq_id] = seq
            new_seq_list = list(id_dict.values())


        elif self.style == 'remove':
            id_dict = {}
            for seq in sequence:
                seq_id = seq.id
                if seq_id not in id_dict:
                    id_dict[seq_id] = seq
            new_seq_list = list(id_dict.values())

        return new_seq_list



'''Returns a list of Sequence objects with the reverse complement of the sequences.'''
# pylint: disable=too-few-public-methods
from packages.sequences.sequences import Sequence
from .abstract_transformer import AbstractTransformer

class ReverseComplement(AbstractTransformer):
    '''Returns a list of Sequence objects with the reverse complement of the sequences.'''
    def __init__(self, style):
        self.style = style

    def transform(self, sequence):
        '''Returns a list of Sequence objects with the specified transformation.'''
        toret = []
        if self.style == 'reverse':
            for seq in sequence:
                toret.append(Sequence(seq.id, seq.seq[::-1], seq.file_name))

        elif self.style == 'complement':
            complements = {
                'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A',
                'a': 't', 'c': 'g', 'g': 'c', 't': 'a'
            }
            for seq in sequence:
                toret.append(Sequence(seq.id, ''.join(complements[base] for base in seq.seq),
                                      seq.file_name))

        elif self.style == 'both':
            complements = {
                'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A',
                'a': 't', 'c': 'g', 'g': 'c', 't': 'a'
            }
            for seq in sequence:
                combined_seq = ''.join(complements[base] for base in seq.seq[::-1])
                toret.append(Sequence(seq.id, combined_seq, seq.file_name))

        return toret



'''Abstract class for sequence transformers.'''
# pylint: disable=too-few-public-methods
from abc import ABC, abstractmethod

class AbstractTransformer(ABC):
    '''Abstract class for sequence transformers.'''
    @abstractmethod
    def transform(self, sequence):
        '''Returns a list of Sequence objects with the transformation applied.'''



'''Module for generating plots from sequences.'''
import os
import pandas as pd
import matplotlib.pyplot as plt

class PlotGenerator:
    '''Class for generating plots from sequences.'''
    def __init__(self, sequences, input_path, plots_dir):
        self.sequences = sequences
        self.input_path = input_path
        self.plots_dir = plots_dir
        self.df = self.create_dataframe()

    def create_dataframe(self):
        '''Creates a DataFrame from the sequences.'''
        seq_ids = [seq.id for seq in self.sequences]
        seq_lengths = [len(seq.seq) for seq in self.sequences]

        base_counts = {'A': [], 'C': [], 'T': [], 'G': []}
        for seq in self.sequences:
            for base, count_list in base_counts.items():
                count_list.append(seq.seq.count(base))

        data = {
            'seq_id': seq_ids,
            'len': seq_lengths,
            'A': base_counts['A'],
            'C': base_counts['C'],
            'T': base_counts['T'],
            'G': base_counts['G']
        }
        return pd.DataFrame(data)

    def generate_histogram(self):
        '''Generates a histogram of sequence lengths.'''
        sequence_lengths = self.df['len']
        plt.hist(sequence_lengths, bins=10, color='#DDA0DD')
        plt.xlabel('Sequence length')
        plt.ylabel('Frequency')
        plt.title('Histogram of sequence length distribution')
        plt.grid(True)
        input_file_name = os.path.splitext(os.path.basename(self.input_path))[0]
        histogram_filename = input_file_name + '_histogram.png'
        histogram_filepath = os.path.join(self.plots_dir, histogram_filename)
        os.makedirs(self.plots_dir, exist_ok=True)
        plt.savefig(histogram_filepath)
        plt.close()

    def generate_boxplot(self):
        '''Generates a box plot of base percentages.'''
        base_counts = ['A', 'C', 'T', 'G']
        total_lengths = self.df['len']
        for base in base_counts:
            self.df[base] = self.df[base] / total_lengths

        box_colors = ['#FFC0CB', '#ADD8E6', '#90EE90', '#FFFF66']
        plt.boxplot(self.df[base_counts].values, labels=base_counts, patch_artist=True)

        plt.xticks([])

        for patch, color in zip(plt.boxplot(self.df[base_counts].values,patch_artist=True)['boxes'],
                                box_colors):
            patch.set_facecolor(color)

        plt.xlabel('Bases')
        plt.ylabel('Percentage')
        plt.title('Relative percentage of bases in the total length of sequences')
        
        plt.grid(True)
        input_file_name = os.path.splitext(os.path.basename(self.input_path))[0]
        boxplot_filename = input_file_name + '_boxplot.png'
        boxplot_filepath = os.path.join(self.plots_dir, boxplot_filename)
        plt.savefig(boxplot_filepath)
        plt.close()

    def generate_plots(self):
        '''Generates plots from the sequences.'''
        self.generate_histogram()
        self.generate_boxplot()



'''Module for loading and writing fasta files.'''
import os
from packages.sequences.sequences import Sequence

class SeqFileManager:
    '''This class allows us to format sequences and load and write fasta files.'''
    def __init__(self, case=0, input_filename=None):
        self.case = case
        self.input_filename = input_filename


    def load_fasta(self, file_name):
        '''Returns a list of Sequence objects from the fasta file.'''
        sequences = []
        with open(file_name, 'r', encoding='utf-8') as file:
            current_id = None
            current_seq = ''
            for line in file:
                line = line.strip()
                if line.startswith('>'):
                    if current_id is not None:
                        # Provide the file_name parameter when creating Sequence objects
                        sequences.append(Sequence(current_id, current_seq, file_name))
                    current_id = line[1:]
                    current_seq = ''
                else:
                    current_seq += line
            if current_id is not None:
                # Provide the file_name parameter when creating Sequence objects
                sequences.append(Sequence(current_id, current_seq, file_name))
        return sequences

    # En la clase SeqFileManager en packages/Utils/seq_file_manager.py
    def write_fasta(self, sequences, output_directory):
        '''Writes all sequences to individual files in the output directory.'''
        os.makedirs(output_directory, exist_ok=True)  # Ensure the output directory exists

        for i, sequence in enumerate(sequences):
            # Use the input file name (sequence.file_name) without extension for the output file
            input_file_name = os.path.splitext(os.path.basename(sequence.file_name))[0]
            output_filename = f"{input_file_name}_result.fasta"
            output_file_path = os.path.join(output_directory, output_filename)

            # Use 'w' mode for the first sequence, 'a' mode for subsequent sequences
            mode = 'w' if i == 0 else 'a'

            with open(output_file_path, mode, encoding='utf-8') as f:
                f.write(f'>{sequence.id}\n')
                f.write(f'{sequence.seq}\n')



'''This module contains the Sequence class.'''
## Definition of the Sequence class

class Sequence:
    '''This class represents a sequence.'''
    def __init__(self, sequence_id, sequence, file_name=None):
        self.id = sequence_id
        self.seq = sequence
        self.file_name = file_name

    def __repr__(self):
        return f'{self.id}: {self.seq}'

    def __str__(self):
        return f'{self.id}: {self.seq}\n'

    def lower(self):
        '''Returns a new Sequence object with the sequence in lowercase.'''
        return Sequence(self.id, self.seq.lower(), self.file_name)

    def upper(self):
        '''Returns a new Sequence object with the sequence in uppercase.'''
        return Sequence(self.id, self.seq.upper(), self.file_name)
